from .edit_resources_tk import EditResourcesTk
from .edit_resources_tk import DEColumnDescription
from .edit_courses_tk import EditCoursesTk
from .main_pages_tk import MainPageBaseTk
# from .scheduler_tk import SchedulerTk, set_main_page_event_handler, MAIN_PAGE_EVENT_HANDLERS
