PREP_FACTOR                = 0.9
PREP_BONUS_LIMIT           = 3.0
PREP_BONUS_FACTOR          = 0.2  # 1.1 - 0.9
PREP_CRAZY_BONUS_FACTOR    = 1.0

PES_FACTOR                 = 0.04
PES_BONUS_LIMIT            = 415
PES_BONUS_FACTOR           = 0.03

STUDENT_BONUS_LIMIT        = 75
STUDENT_BONUS_FACTOR       = 0.01
STUDENT_CRAZY_BONUS_LIMIT  = 160
STUDENT_CRAZY_BONUS_FACTOR = 0.1

HOURS_FACTOR               = 1.2

CI_FTE_PER_SEMESTER        = 40
